#include "mv.hh"

#include <iostream>
#include <stdlib.h> // for rand() function

using namespace std ;

// initialize structure to a matrix of size 0 x 0
void
Mat::Initialize( ) {
  this->nrow         = 0 ;
  this->ncol         = 0 ;
  this->memory_block = nullptr ;
  this->values       = nullptr ;
}

// alocate a matrix to the size: nrow x ncol
void
Mat::Allocate( int nrow, int ncol ) { // USE this to distinguish arguments
  // (*this).nrow ==> this -> nrow
  if ( this->nrow == nrow &&  this->ncol == ncol ) return ; // nothing to do!

  this->nrow = nrow ;
  this->ncol = ncol ;

  if ( this->memory_block != nullptr ) delete [] this->memory_block ;
  this->memory_block = new double [nrow*ncol] ;

  if ( this->values != nullptr ) delete [] this->values ;
  this->values = new double* [this->nrow] ;

  // initialize the pointer of pointer
  for ( int i = 0 ; i < this->nrow ; ++i )
    this->values[i] = this->memory_block + i * this->ncol ;
}

void
Mat::AllocateAndInitializeRandom( int nrow, int ncol ) {
  this->Allocate( nrow, ncol ) ;
  for ( int i = 0 ; i < this->nrow ; ++i )
    for ( int j = 0 ; j < this->ncol ; ++j )
      this->values[i][j] = rand() - rand() ;
}

void
Mat::Fill( double val ) {
  for ( int i = 0 ; i < this->nrow ; ++i )
    for ( int j = 0 ; j < this->ncol ; ++j )
      this->values[i][j] = val ;
}

void
Mat_x_Vec( struct Mat const & mat, // ensure the compiler that mat is not modified inside...
           struct Vec const & vec,
           struct Vec       & res ) {
  // check matrix vector compatibility
  if ( mat.ncol != vec.size ) {
    cerr << "Mat_x_Vec, incompatible dimensions\n" ;
    exit(0) ;
  }   
  if ( res.size != mat.nrow ) res.Allocate( mat.nrow ) ;
  for ( int i = 0 ; i < mat.nrow ; ++i ) {
    res(i) = 0 ;
    for ( int j = 0 ; j < mat.ncol ; ++i )
      res(i) += mat(i,j)*vec(j) ;
  }
}

void
Mat_x_Mat_123( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_123, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.Allocate( A.nrow, B.ncol ) ;

  // set to 0
  C.Fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int i = 0 ; i < A.nrow ; ++i ) // 1
    for ( int j = 0 ; j < B.ncol ; ++j ) // 2
      for ( int k = 0 ; k < A.ncol ; ++k ) // 3
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_132( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_132, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.Allocate( A.nrow, B.ncol ) ;

  // set to 0
  C.Fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int i = 0 ; i < A.nrow ; ++i ) // 1
    for ( int k = 0 ; k < A.ncol ; ++k ) // 3
      for ( int j = 0 ; j < B.ncol ; ++j ) // 2
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_213( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_213, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.Allocate( A.nrow, B.ncol ) ;

  // set to 0
  C.Fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int j = 0 ; j < B.ncol ; ++j ) // 2
    for ( int i = 0 ; i < A.nrow ; ++i ) // 1
      for ( int k = 0 ; k < A.ncol ; ++k ) // 3
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_231( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_231, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.Allocate( A.nrow, B.ncol ) ;

  // set to 0
  C.Fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int j = 0 ; j < B.ncol ; ++j ) // 2
    for ( int k = 0 ; k < A.ncol ; ++k ) // 3
      for ( int i = 0 ; i < A.nrow ; ++i ) // 1
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_312( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_312, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.Allocate( A.nrow, B.ncol ) ;

  // set to 0
  C.Fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int k = 0 ; k < A.ncol ; ++k ) // 3
    for ( int i = 0 ; i < A.nrow ; ++i ) // 1
      for ( int j = 0 ; j < B.ncol ; ++j ) // 2
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_321( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_321, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.Allocate( A.nrow, B.ncol ) ;

  // set to 0
  C.Fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int k = 0 ; k < A.ncol ; ++k ) // 3
    for ( int j = 0 ; j < B.ncol ; ++j ) // 2
      for ( int i = 0 ; i < A.nrow ; ++i ) // 1
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_with_tiling( struct Mat const & A, 
                       struct Mat const & B,
                       struct Mat       & C,
                       int tsize ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_321, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.Allocate( A.nrow, B.ncol ) ;

  // set to 0
  C.Fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int ii = 0 ; ii < A.nrow ; ii += tsize ) { // 1
    int ii_max = std::min(ii + tsize, A.nrow) ;
    for ( int kk = 0 ; kk < A.ncol ; kk += tsize ) { // 3
      int kk_max = std::min(kk + tsize, A.ncol) ;
      for ( int jj = 0 ; jj < B.ncol ; jj += tsize ) { // 2
        int jj_max = std::min(jj + tsize, B.ncol) ;
        for ( int i = ii ; i < ii_max ; ++i ) // 1
          for ( int k = kk ; k < kk_max ; ++k ) // 3
            for ( int j = jj ; j < jj_max ; ++j ) // 2
              C(i,j) += A(i,k)*B(k,j) ;
      }
    }
  }
}

void
Mat::Free() {
  // check if alreay de-allocated
  if ( this->memory_block != nullptr ) delete [] this->memory_block ;
  if ( this->values       != nullptr ) delete [] this->values ;
  this->nrow = 0 ;
  this->ncol = 0 ;
}
